cc.wsm = {current_id = 1} --inject websocket manager into cc.wsm

--websocket states
WS_READYSTATE_CONNECTING = 0
WS_READYSTATE_OPEN = 1
WS_READYSTATE_CLOSING = 2
WS_READYSTATE_CLOSED = 3
WS_READYSTATE_ERROR = 4

WS_PROXY_COMMAND_OPEN = '1'
WS_PROXY_COMMAND_CLOSE = '3'
WS_PROXY_COMMAND_ERROR = '4'
WS_PROXY_COMMAND_MSG = '2'

--WS_DATA_FORMAT_TEXT = 0
--WS_DATA_FORMAT_BINARY = 1

--local websock = {}
local websock = class('WebSocket')
local TIMEOUT_EVENT_ID =  2008
local WEBSOCKET_TIMEOUT = 10

local TAG = '[WebSocketProxy]'

local function get_connect_id()
    local cid = cc.wsm.current_id
    cc.wsm.current_id = cc.wsm.current_id + 1
    return cid
end

local function splitString(s, p)
    local subStrings= {}
    string.gsub(
        s ,
        '[^'..p..']+' ,
        function(w)
            table.insert(subStrings, w) 
        end
    )
    return subStrings
end

local function trim(s)
  return string.match(s,'^()%s*$') and '' or string.match(s,'^%s*(.*%S)')
end

local function parseProtocol(url)
    local subStrings = splitString(url,':')
    if #subStrings ~= 2 then
        return 'ws'
    else 
        return subStrings[1] 
    end
end

function websock:ctor(url)
    local ws = {}
    setmetatable(ws, self)  
    self.__index = self

    ws.url = url
    ws.protocol = parseProtocol(url)
    ws.readystate = WS_READYSTATE_CONNECTING
    ws.event_handle = {}
    ws.cid = get_connect_id()
    ws.timeID = nil
    
    ws.proxy_send_text_header = '[2,'..tostring(ws.cid)..',1]'
    ws.proxy_send_binary_header = '[2,'..tostring(ws.cid)..',2]'

    ws.dataType = WS_DATA_FORMAT_TEXT
    ws.proxy_send_header = ws.proxy_send_text_header

    cc.wsm[ws.cid] = ws    
end

function websock:create(url)
    local ws = {}
    setmetatable(ws, self)  
    self.__index = self    

    ws.url = url
    ws.protocol = parseProtocol(url)
    ws.readystate = WS_READYSTATE_CONNECTING
    ws.event_handle = {}
    ws.cid = get_connect_id()

    ws.proxy_send_text_header = '[2,'..tostring(ws.cid)..',1]'
    ws.proxy_send_binary_header = '[2,'..tostring(ws.cid)..',2]'

    ws.dataType = WS_DATA_FORMAT_TEXT
    ws.proxy_send_header = ws.proxy_send_text_header

    cc.wsm[ws.cid] = ws    
    ws:connect()

    return ws
end

function websock:setDataType(dataType)
    print('websock:setDataType ,datatype is '..tostring(dataType))
    self.dataType = dataType
    if self.dataType == WS_DATA_FORMAT_BINARY then
        self.proxy_send_header = self.proxy_send_binary_header
    else
        self.proxy_send_header = self.proxy_send_text_header
    end
end

function websock:connect()    
    cc.NativeBridge:getInstance():gameSendData(
        '' ,
        '[1,'..tostring(self.cid)..',"'..self.url..'"]' ,
        ''
    )
    self.timeID = cc.Director:getInstance():getScheduler():scheduleScriptFunc(
        handler(
            cid ,
            function(cid)
                local ws = cc.wsm[cid]
                if ws ~= nil and self.timeID~= nil then                    
                    ws:onSocketState(cc.WEBSOCKET_ERROR)
                    if cc.Director:getInstance() ~= nil then
                        cc.Director:getInstance():getScheduler():unscheduleScriptEntry(ws.timeID)
                    end
                    ws.timeID = nil             
                end
            end
        ),
        WEBSOCKET_TIMEOUT ,
        false
    )
end

function websock:setNotAutoReConnect(notAutoRectConnect)

end

function websock:getReadyState()
    return self.readystate
end

function websock:close()
    if cc.wsm[self.cid] == nil then
        return 
    end

    cc.NativeBridge:getInstance():gameSendData(
        '' ,
        '[3,'..tostring(self.cid)..']' ,
        ''
    )

    self:clearupTimer()
    cc.wsm[self.cid] = nil

    self:onSocketState(cc.WEBSOCKET_CLOSE)
end

function websock:sendString(data)    
    cc.NativeBridge:getInstance():gameSendData(
        '' ,
        self.proxy_send_header ,
        data
    )
end

function websock:clearupTimer()
    if self.timeID ~= nil then
        if cc.Director:getInstance() ~= nil then
            cc.Director:getInstance():getScheduler():unscheduleScriptEntry(self.timeID)
        end
        self.timeID = nil --cancel timer
    end    
end

function websock:registerScriptHandler(func,evt)
    --print(string.format('%sregisterScriptHandler evt=%d,func=%s',TAG,evt,func))
    self.event_handle[evt] = func
end

function websock:unregisterScriptHandler(evt)
    print(string.format('%unregisterScriptHandler evt=%d',TAG,evt))
    self.event_handle[evt] = nil    
end

function websock:onSocketState(evt)
    print(TAG..'websock:onSocketState connectionID='..self.cid..',evt='..evt)
    if evt == cc.WEBSOCKET_OPEN then        
        self.readystate = WS_READYSTATE_OPEN
        print(TAG..'websock:onSocketState ，websocket opened ,connectionID='..self.cid)
    end

    local func = self.event_handle[evt]
    if func ~= nil and type(func)=='function' then
        func('')       
    end

    self:clearupTimer()
    if evt == cc.WEBSOCKET_CLOSE then
        self.event_handle={}        
        cc.wsm[self.cid] = nil        
        self.readystate = WS_READYSTATE_CLOSED
        print(TAG..'websock:onSocketState ，websocket closed ,connectionID='..self.cid)
    elseif evt == cc.WEBSOCKET_ERROR then
        self.readystate = WS_READYSTATE_ERROR
        print(TAG..'websock:onSocketState ，websocket error ,connectionID='..self.cid)
    --elseif evt == cc.WEBSOCKET_OPEN then        
    --    self.readystate = WS_READYSTATE_OPEN
    --    print(TAG..'websock:onSocketState ，websocket opened ,connectionID='..self.cid)
    end
end

local function parseHeader(header)
    if string.len(header) < 2 then
        print(TAG..'invalid header ,header length is less than 2 .header='..header)
        return nil
    end

    local headerTable = splitString(string.sub(header ,2 ,string.len(header)-1) ,',')    
    if #headerTable < 2 or #headerTable > 3 then
        print(TAG..'invalid header ,format error ,header='..header)
        return nil        
    elseif #headerTable > 2 then
        headerTable.opcode = trim(headerTable[3]) --tonumber(headerTable[3])
    end

    headerTable.command = trim(headerTable[1]) --tonumber(headerTable[1])
    headerTable.connectionID = tonumber(headerTable[2])

    return headerTable
end

function appReceiveData(context ,header ,body)
    --print(TAG..'appReceiveData ,header='..header..',body='..body)
    local headerTable = parseHeader(header)
    if headerTable == nil then
        print(TAG..'appReceiveData ,invalid header ,header format error'..header)
        return
    end

    local ws = cc.wsm[headerTable.connectionID]
    if ws == nil then
        print(TAG..'appReceiveData ,connection id is closed or invalid,connectionId='..headerTable.connectionID)
        return
    end    
    
    if headerTable.command == WS_PROXY_COMMAND_MSG then
        local msgHandler = ws.event_handle[cc.WEBSOCKET_MESSAGE]
        if msgHandler ~= nil then
            --print(TAG..'appReceiveData onGameMessage ,msg='..body)
            msgHandler(body)
        else
            print(TAG..'appReceiveData ，receive message but no handler,header='..header)
        end
    elseif headerTable.command == WS_PROXY_COMMAND_CLOSE then       
        ws:onSocketState(cc.WEBSOCKET_CLOSE)        
    elseif headerTable.command == WS_PROXY_COMMAND_OPEN then          
        ws:onSocketState(cc.WEBSOCKET_OPEN)
    elseif headerTable.command == WS_PROXY_COMMAND_ERROR then
        ws:onSocketState(cc.WEBSOCKET_ERROR)
    end    
end

function appChannelStateChanged(context ,seqid ,state)
    local state = string.byte(state)
    if state == cc.APP_NET_CHANNEL_INIT 
        or state == APP_NET_CHANNEL_CONNECTING 
        or state == APP_NET_CHANNEL_RECONNECTING 
        or state == APP_NET_CHANNEL_CONNECTED then
        print(TAG..'appchannel state changed for init or connecting/reconnecting')
    elseif state == cc.APP_NET_CHANNEL_READY then               
        for _ ,ws in pairs(cc.wsm) do
            if ws ~=nil and type(ws)~= 'number' then
                ws:connect()
            end            
        end
        print(TAG..'appchannel state changed ,state = APP_NET_CHANNEL_READY')
    elseif state == cc.APP_NET_CHANNEL_ERROR or state == cc.APP_NET_CHANNEL_LINKERROR then        
        for _ ,ws in pairs(cc.wsm) do            
            if ws ~=nil and type(ws)~= 'number' then
                ws:onSocketState(cc.WEBSOCKET_ERROR)
            end            
        end
        print(TAG..'appchannel state changed ,state = WEBSOCKET_ERROR')
    end    
end

return websock